#define _GNU_SOURCE

#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/syscall.h>
#include <unistd.h>

#include <fcntl.h>
#include <stdarg.h>
#include <dlfcn.h>
#include <libgen.h>
#include <limits.h>
#include <errno.h>
#include <paths.h>

#include <sys/utsname.h>
#include <dirent.h>
#include <sys/stat.h>

/*
 * execve wrapping is taken directly from fakeroot
 * please look at fakeroot source code for examples, documentation
 * and overall understanding how this works :-)
 * */


#define LIBCPATH "/scratchbox/host_shared/lib/libc.so.6"
#define SCRATCHBOX_TOOLS_BIN "/scratchbox/tools/bin"

extern char **environ;

struct next_wrap_st {
	void **doit;
	char *name;
};

void *get_libc()
{
#ifndef RTLD_NEXT
	void *lib=0;
	if (!lib) {
		lib = dlopen(LIBCPATH, RTLD_LAZY);
	}
	if (NULL == lib) {
		fprintf(stderr, "Couldn't find libc at: %s\n", LIBCPATH);
		abort();
	}
	return lib;
#else
	return RTLD_NEXT;
#endif
}


void load_library_symbols(void);

extern int (*next_execve)(const char *file, char *const argv[], char *const envp[]);
int tmp_execve(const char *file, char *const argv[], char *const envp[]) {
	load_library_symbols();
	return next_execve(file,argv,envp);
}

int (*next_execve)(const char *file, char *const argv[], char *const envp[])=tmp_execve;

struct next_wrap_st next_wrap[]= {
	{(void(*))&next_execve, "execve"},
	{NULL, NULL}
};

void load_library_symbols(void)
{
	static int done=0;
	int i;
	char *msg;

	if (!done) {
		for (i=0; next_wrap[i].doit; i++) {
			*(next_wrap[i].doit)=dlsym(get_libc(), next_wrap[i].name);
			if ( (msg = dlerror()) != NULL) {
				fprintf (stderr, "dlsym(%s): %s\n", next_wrap[i].name, msg);
			}
		}
	}
}


int check_for_hashbang(const char *file, char *interpreter, char *interpreter_arg)
{
	int hash_bang = 0, i = 0, j = 0, c=0;
	char buf[1024]; // hash bang detection buffer
	FILE *fp;

	memset(buf, 0, 1024);

	if ((fp = fopen(file, "r"))) {
		c = fread(buf, 1023, 1, fp);
		if (buf[0] == '#' && buf[1] == '!') {
			// we have hash bang
			//fprintf(stderr, "we have a hash bang!\n");
			i = 2;
			while ((buf[i] == ' ' || buf[i] == '\t') && buf[i] != '\0' && buf[i] != '\n') i++;
			j = i;
			while (buf[j] != ' ' && buf[j] != '\t' && buf[j] != '\0' && buf[j] != '\n') j++;
			strncpy(interpreter, &buf[i], j - i);
			while ((buf[j] == ' ' || buf[j] == '\t') && buf[j] != '\0' && buf[j] != '\n') j++;
			i = j;
			while (buf[j] != '\n' && buf[j] != '\0') j++;
			strncpy(interpreter_arg, &buf[i], j - i);
			//fprintf(stderr, "interpreter: [%s]\n", interpreter);
			//fprintf(stderr, "arg: [%s]\n", interpreter_arg);
			hash_bang = 1;
		}
		fclose(fp);
	}

	return hash_bang;
}

int prepare_hashbang_call(const char *file, char *const argv[], char *interpreter, char *interpreter_arg, char *real_file, char ***real_argv)
{
	char **tmp;
	char **tmp2;
	int c=0;
	
	realpath(interpreter, real_file);
	tmp = (char **)argv;
	c = 0;
	while (*tmp) {
		c++;
		tmp++;
	}
	*real_argv = malloc(sizeof(char **)* (c + 3));
	tmp = (char **)argv;
	tmp2 = (char **)*real_argv;
	*tmp2++ = (char *)file;
	if (interpreter_arg[0] != '\0') {
		*tmp2++ = interpreter_arg;
	}

	*tmp2++ = (char *)file;

	tmp++; // skip argv[0]
	while (*tmp) {
		*tmp2++ = *tmp++;
	}
	*tmp2 = NULL;
	return 1;
}
		
int execve (file, argv, envp)
	const char *file;
	char *const argv[];
	char *const envp[];
{
	char *dir, *bin, *redir_list, *li, **my_argv;
	char *tmp_file;
	char real_file[PATH_MAX];
	char real_thing[PATH_MAX];
	char interpreter[1024];
	char interpreter_arg[1024];

	int hash_bang=0;
	DIR *stb;
	struct dirent *d;
	
	fprintf(stderr, "fake execve!\n");

	realpath(file, real_file);

	//fprintf(stderr, "TSS: %s\n", real_file);
	
	if (!(redir_list = getenv("SBOX_REDIRECT_BINARIES"))) {
		/* default as there is nothing to redirect */
		return next_execve(file, argv, envp);
	}


	li = strdup(redir_list);
	redir_list = li;
	li = NULL;

	memset(interpreter, 0, 1024);
	hash_bang = check_for_hashbang(file, interpreter, interpreter_arg);

	if (hash_bang) {
		prepare_hashbang_call(file, argv, interpreter, interpreter_arg, tmp_file, &my_argv);
	} else {
		my_argv = (char **)argv;
		tmp_file = (char *)real_file;
	}

	dir = dirname(strdup(tmp_file));
	bin = basename(strdup(tmp_file));

	while ((li = strsep(&redir_list, "#"))) {
		char *exe1 = NULL;
		char *exe2 = NULL;

		exe1 = strsep(&li, ":");
		exe2 = strsep(&li, ":");

		if (!exe1 || !exe2) continue;

		realpath(exe1,real_thing);
		if ( (strcmp(tmp_file, real_thing) == 0) ) {
			return next_execve(exe2, my_argv, envp);
		}
	}

	if (getenv("SBOX_REDIRECT_BIN") && strcmp(dirname(strdup(file)), "/bin") == 0) {
		// we want to redirect all binaries in /scratchbox/tools/bin
		// that are executed from /bin
		fprintf(stderr, "redirecting /bin\n");
		stb = opendir(SCRATCHBOX_TOOLS_BIN);
		while ((d = readdir(stb))) {
			// loop through the files 
			struct stat f;
			char fullname[strlen(SCRATCHBOX_TOOLS_BIN) + strlen(d->d_name) + 2];
			fullname[0] = '\0';
			strcat(fullname, SCRATCHBOX_TOOLS_BIN);
			strcat(fullname, "/");
			strcat(fullname, d->d_name);
			fprintf(stderr, "looping: %s\n", fullname);
			if (stat(fullname, &f) != 0) {
				perror("ERROR");
				continue; // error stat:ing the file
			}
			fprintf(stderr, "stat OK\n");
			if (!S_ISREG(f.st_mode)) continue; // not a regular file
			fprintf(stderr, "matching: [%s]:[%s]\n", d->d_name, bin);
			if (strcmp(d->d_name, bin) == 0) {
				// found a match
				char *exe = malloc(strlen(SCRATCHBOX_TOOLS_BIN) + strlen(bin) + 2);
				*exe = '\0';
				strcat(exe,SCRATCHBOX_TOOLS_BIN);
				strcat(exe,"/");
				strcat(exe,bin);
				fprintf(stderr, "found a match: %s\n", exe);
				memset(interpreter, 0, 1024);
				hash_bang = check_for_hashbang(exe, interpreter, interpreter_arg);

				if (hash_bang) {
					prepare_hashbang_call(exe, argv, interpreter, interpreter_arg, tmp_file, &my_argv);
				} else {
					my_argv = (char **)argv;
					tmp_file = (char *)exe;
				}

				return next_execve(tmp_file, my_argv, envp);
			}
		}
	}

	/* fall back to default behaviour */
	
	//fprintf(stderr, "default handler\n");
	return next_execve(file, argv, envp);
}

int uname(struct utsname *buf) 
{
	char *my_sysname, *my_nodename, *my_release, *my_version, *my_machine;
	if ((my_sysname = getenv("SBOX_UNAME_SYSNAME"))) {
		strncpy(buf->sysname, my_sysname, _UTSNAME_LENGTH);
	} else {
		strncpy(buf->sysname,"Linux", _UTSNAME_LENGTH);
	}

	if ((my_nodename = getenv("SBOX_UNAME_NODENAME"))) {
		strncpy(buf->nodename, my_nodename, _UTSNAME_LENGTH);
	} else {
		strncpy(buf->nodename, "fake-node", _UTSNAME_LENGTH);
	}

	if ((my_release = getenv("SBOX_UNAME_RELEASE"))) {
		strncpy(buf->release, my_release, _UTSNAME_LENGTH);
	} else {
		strncpy(buf->release, "2.4.21", _UTSNAME_LENGTH);
	}

	if ((my_version = getenv("SBOX_UNAME_VERSION"))) {
		strncpy(buf->version, my_version, _UTSNAME_LENGTH);
	} else {
		strncpy(buf->version, "#1 Sun Jul 20 19:23:36 EST 2003", _UTSNAME_LENGTH);
	}

	if ((my_machine = getenv("SBOX_UNAME_MACHINE"))) {
		strncpy(buf->machine, my_machine, _UTSNAME_LENGTH);
	} else {
		strncpy(buf->machine, "i386", _UTSNAME_LENGTH);
	}
	return 0;
}






/* ************ exec family of functions copied from glibc-2.3.2 ***********/


#ifndef	HAVE_GNU_LD
# define __environ	environ
#endif

/* Execute PATH with all arguments after PATH until
   a NULL pointer and environment from `environ'.  */
int
execl (const char *path, const char *arg, ...)
{
  size_t argv_max = 1024;
  const char **argv = alloca (argv_max * sizeof (const char *));
  unsigned int i;
  va_list args;

  argv[0] = arg;

  va_start (args, arg);
  i = 0;
  while (argv[i++] != NULL)
    {
      if (i == argv_max)
	{
	  const char **nptr = alloca ((argv_max *= 2) * sizeof (const char *));

#ifndef _STACK_GROWS_UP
	  if ((char *) nptr + argv_max == (char *) argv)
	    {
	      /* Stack grows down.  */
	      argv = (const char **) memcpy (nptr, argv,
					     i * sizeof (const char *));
	      argv_max += i;
	    }
	  else
#endif
#ifndef _STACK_GROWS_DOWN
	    if ((char *) argv + i == (char *) nptr)
	    /* Stack grows up.  */
	    argv_max += i;
	  else
#endif
	    /* We have a hole in the stack.  */
	    argv = (const char **) memcpy (nptr, argv,
					   i * sizeof (const char *));
	}

      argv[i] = va_arg (args, const char *);
    }
  va_end (args);

  return execve (path, (char *const *) argv, __environ);
}




/* Execute PATH with all arguments after PATH until a NULL pointer,
   and the argument after that for environment.  */
int
execle (const char *path, const char *arg, ...)
{
  size_t argv_max = 1024;
  const char **argv = alloca (argv_max * sizeof (const char *));
  const char *const *envp;
  unsigned int i;
  va_list args;
  argv[0] = arg;

  va_start (args, arg);
  i = 0;
  while (argv[i++] != NULL)
    {
      if (i == argv_max)
	{
	  const char **nptr = alloca ((argv_max *= 2) * sizeof (const char *));

#ifndef _STACK_GROWS_UP
	  if ((char *) nptr + argv_max == (char *) argv)
	    {
	      /* Stack grows down.  */
	      argv = (const char **) memcpy (nptr, argv,
					     i * sizeof (const char *));
	      argv_max += i;
	    }
	  else
#endif
#ifndef _STACK_GROWS_DOWN
	    if ((char *) argv + i == (char *) nptr)
	    /* Stack grows up.  */
	    argv_max += i;
	  else
#endif
	    /* We have a hole in the stack.  */
	    argv = (const char **) memcpy (nptr, argv,
					   i * sizeof (const char *));
	}

      argv[i] = va_arg (args, const char *);
    }

  envp = va_arg (args, const char *const *);
  va_end (args);

  return execve (path, (char *const *) argv, (char *const *) envp);
}




/* Execute FILE, searching in the `PATH' environment variable if
   it contains no slashes, with all arguments after FILE until a
   NULL pointer and environment from `environ'.  */
int
execlp (const char *file, const char *arg, ...)
{
  size_t argv_max = 1024;
  const char **argv = alloca (argv_max * sizeof (const char *));
  unsigned int i;
  va_list args;

  argv[0] = arg;

  va_start (args, arg);
  i = 0;
  while (argv[i++] != NULL)
    {
      if (i == argv_max)
	{
	  const char **nptr = alloca ((argv_max *= 2) * sizeof (const char *));

#ifndef _STACK_GROWS_UP
	  if ((char *) nptr + argv_max == (char *) argv)
	    {
	      /* Stack grows down.  */
	      argv = (const char **) memcpy (nptr, argv,
					     i * sizeof (const char *));
	      argv_max += i;
	    }
	  else
#endif
#ifndef _STACK_GROWS_DOWN
	    if ((char *) argv + i == (char *) nptr)
	    /* Stack grows up.  */
	    argv_max += i;
	  else
#endif
	    /* We have a hole in the stack.  */
	    argv = (const char **) memcpy (nptr, argv,
					   i * sizeof (const char *));
	}

      argv[i] = va_arg (args, const char *);
    }
  va_end (args);

  return execvp (file, (char *const *) argv);
}


#ifndef	HAVE_GNU_LD
# define __environ	environ
#endif

/* Execute PATH with arguments ARGV and environment from `environ'.  */
int
execv (const char *path, char *const argv[])
{
  return execve (path, argv, __environ);
}




/* The file is accessible but it is not an executable file.  Invoke
   the shell to interpret it as a script.  */
static void
sb_script_execute (const char *file, char *const argv[])
{
  /* Count the arguments.  */
  int argc = 0;
  while (argv[argc++])
    ;

  /* Construct an argument list for the shell.  */
  {
    char *new_argv[argc + 1];
    new_argv[0] = (char *) _PATH_BSHELL;
    new_argv[1] = (char *) file;
    while (argc > 1)
      {
	new_argv[argc] = argv[argc - 1];
	--argc;
      }

    /* Execute the shell.  */
    execve (new_argv[0], new_argv, __environ);
  }
}


/* Execute FILE, searching in the `PATH' environment variable if it contains
   no slashes, with arguments ARGV and environment from `environ'.  */
int
execvp (file, argv)
     const char *file;
     char *const argv[];
{
  if (*file == '\0')
    {
      /* We check the simple case first. */
      errno = ENOENT;
      return -1;
    }

  if (strchr (file, '/') != NULL)
    {
      /* Don't search when it contains a slash.  */
      execve (file, argv, __environ);

      if (errno == ENOEXEC)
	sb_script_execute (file, argv);
    }
  else
    {
      int got_eacces = 0;
      char *path, *p, *name;
      size_t len;
      size_t pathlen;

      path = getenv ("PATH");
      if (path == NULL)
	{
	  /* There is no `PATH' in the environment.
	     The default search path is the current directory
	     followed by the path `confstr' returns for `_CS_PATH'.  */
	  len = confstr (_CS_PATH, (char *) NULL, 0);
	  path = (char *) alloca (1 + len);
	  path[0] = ':';
	  (void) confstr (_CS_PATH, path + 1, len);
	}

      len = strlen (file) + 1;
      pathlen = strlen (path);
      name = alloca (pathlen + len + 1);
      /* Copy the file name at the top.  */
      name = (char *) memcpy (name + pathlen + 1, file, len);
      /* And add the slash.  */
      *--name = '/';

      p = path;
      do
	{
	  char *startp;

	  path = p;
	  p = strchrnul (path, ':');

	  if (p == path)
	    /* Two adjacent colons, or a colon at the beginning or the end
	       of `PATH' means to search the current directory.  */
	    startp = name + 1;
	  else
	    startp = (char *) memcpy (name - (p - path), path, p - path);

	  /* Try to execute this name.  If it works, execv will not return.  */
	  execve (startp, argv, __environ);

	  if (errno == ENOEXEC)
	    sb_script_execute (startp, argv);

	  switch (errno)
	    {
	    case EACCES:
	      /* Record the we got a `Permission denied' error.  If we end
		 up finding no executable we can use, we want to diagnose
		 that we did find one but were denied access.  */
	      got_eacces = 1;
	    case ENOENT:
	    case ESTALE:
	    case ENOTDIR:
	      /* Those errors indicate the file is missing or not executable
		 by us, in which case we want to just try the next path
		 directory.  */
	      break;

	    default:
	      /* Some other error means we found an executable file, but
		 something went wrong executing it; return the error to our
		 caller.  */
	      return -1;
	    }
	}
      while (*p++ != '\0');

      /* We tried every element and none of them worked.  */
      if (got_eacces)
	/* At least one failure was due to permissions, so report that
           error.  */
	errno = EACCES;
    }

  /* Return the error from the last attempt (probably ENOENT).  */
  return -1;
}
